@@ -1,5 +1,5 @@
 /* list.c - twoftpd routines to handle list/nlst commands
- * Copyright (C) 2001-2002  Bruce Guenter <bruceg@em.ca>
+ * Copyright (C) 2008  Bruce Guenter <bruce@untroubled.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,12 +24,12 @@
 #include <sysdeps.h>
 #include "twoftpd.h"
 #include "backend.h"
-#include <str/str.h>
+#include <net/socket.h>
 #include <path/path.h>
+#include <str/str.h>
 
 int list_options;
 
-static obuf out;
 static int list_long;
 static int list_flags;
 static int mode_nlst;
@@ -63,12 +63,12 @@ static const char* mode2str(int mode)
   return buf;
 }
 
-static int output_mode(int mode)
+static int str_catmode(str* s, int mode)
 {
-  return obuf_write(&out, mode2str(mode), 10);
+  return str_catb(s, mode2str(mode), 10);
 }
 
-static int output_owner(uid_t owner)
+static int str_catowner(str* s, uid_t owner)
 {
   const char* name;
   unsigned len;
@@ -76,13 +76,13 @@ static int output_owner(uid_t owner)
     name = user, len = user_len;
   else
     name = "somebody", len = 8;
-  if (!obuf_write(&out, name, len)) return 0;
+  if (!str_catb(s, name, len)) return 0;
   while (len++ < MAX_NAME_LEN)
-    if (!obuf_putc(&out, SPACE)) return 0;
+    if (!str_catc(s, SPACE)) return 0;
   return 1;
 }
 
-static int output_group(gid_t g)
+static int str_catgroup(str* s, gid_t g)
 {
   const char* name;
   unsigned len;
@@ -90,13 +90,13 @@ static int output_group(gid_t g)
     name = group, len = group_len;
   else
     name = "somegrp", len = 7;
-  if (!obuf_write(&out, name, len)) return 0;
+  if (!str_catb(s, name, len)) return 0;
   while (len++ < MAX_NAME_LEN)
-    if (!obuf_putc(&out, SPACE)) return 0;
+    if (!str_catc(s, SPACE)) return 0;
   return 1;
 }
 
-static int output_time(time_t then)
+static int str_cattime(str* s, time_t then)
 {
   struct tm* tm;
   time_t year;
@@ -110,73 +110,97 @@ static int output_time(time_t then)
     len = strftime(buf, sizeof buf - 1, "%b %e %H:%M", tm);
   else
     len = strftime(buf, sizeof buf - 1, "%b %e  %Y", tm);
-  return obuf_write(&out, buf, len);
+  return str_catb(s, buf, len);
 }
 
-static int output_stat(const struct stat* s)
+static int str_catstat(str* s, const struct stat* st)
 {
-  if (s)
-    return output_mode(s->st_mode) &&
-      obuf_write(&out, "    1 ", 6) &&
-      output_owner(s->st_uid) &&
-      obuf_putc(&out, SPACE) &&
-      output_group(s->st_gid) &&
-      obuf_putc(&out, SPACE) &&
-      obuf_putuw(&out, s->st_size, 8, ' ') &&
-      obuf_putc(&out, SPACE) &&
-      output_time(s->st_mtime) &&
-      obuf_putc(&out, SPACE);
+  if (st)
+    return str_catmode(s, st->st_mode) &&
+      str_catb(s, "    1 ", 6) &&
+      str_catowner(s, st->st_uid) &&
+      str_catc(s, SPACE) &&
+      str_catgroup(s, st->st_gid) &&
+      str_catc(s, SPACE) &&
+      str_catuwll(s, st->st_size, 8, ' ') &&
+      str_catc(s, SPACE) &&
+      str_cattime(s, st->st_mtime) &&
+      str_catc(s, SPACE);
   else
-    return obuf_puts(&out,
-		     "??????????"
-		     "    ? "
-		     "???????? "
-		     "???????? "
-		     "       ? "
-		     "??? ?? ????? ");
+    return str_cats(s,
+		    "??????????"
+		    "    ? "
+		    "???????? "
+		    "???????? "
+		    "       ? "
+		    "??? ?? ????? ");
 }
 
-static int output_flags(const struct stat* s)
+static int str_catflags(str* s, const struct stat* st)
 {
-  if (s) {
+  if (st) {
     int ch;
-    int mode = s->st_mode;
+    int mode = st->st_mode;
     ch = 0;
     if (S_ISDIR(mode)) ch = '/';
     else if (S_ISFIFO(mode)) ch = '|';
     else if (S_ISSOCK(mode)) ch = '=';
     else if (list_flags == 'F' && mode & 0111) ch = '*';
-    if (ch) return obuf_putc(&out, ch);
+    if (ch)
+      return str_catc(s, ch);
   }
   return 1;
 }
 
-static int obuf_putfn(obuf* o, const char* fn, int striplen)
+static int str_catfn(str* s, const char* fn, int striplen)
 {
   if (striplen < 0) {
-    if (!obuf_putc(o, '/')) return 0;
+    if (!str_catc(s, '/')) return 0;
     striplen = 0;
   }
-  return obuf_puts(o, fn + striplen);
+  return str_cats(s, fn + striplen);
+}
+
+static char send_buf[8192];
+static unsigned long send_used;
+
+static int send_line(int out, const str* s)
+{
+  int result;
+
+  if (send_used + s->len > sizeof send_buf) {
+    if ((result = netwrite(out, send_buf, send_used, timeout * 1000)) != 0)
+      return result;
+    send_used = 0;
+  }
+  if (s->len > sizeof send_buf)
+    return netwrite(out, s->s, s->len, timeout * 1000);
+  memcpy(send_buf + send_used, s->s, s->len);
+  send_used += s->len;
+  return 0;
 }
 
 static str entries;
 
 static int list_entries(long count, int striplen)
 {
+  static str line;
+  int out;
+  int result;
+  unsigned long bytes_out;
   struct stat statbuf;
   struct stat* statptr;
-  int result;
   const char* filename = entries.s;
   int need_stat = list_long || list_flags;
   
   if (mode_nlst && count < 0)
     return respond(550, 1, "No such file or directory");
   
-  if (!make_out_connection(&out)) return 1;
+  if ((out = make_out_connection()) == -1)
+    return 1;
 
-  for (; count > 0; --count, filename += strlen(filename)+1) {
-    result = 1;
+  send_used = 0;
+  for (bytes_out = 0; count > 0; --count, filename += strlen(filename)+1) {
     statptr = 0;
     if (need_stat) {
       if (stat(filename, &statbuf) == -1) {
@@ -185,34 +209,65 @@ static int list_entries(long count, int striplen)
       else
 	statptr = &statbuf;
     }
-    if (list_long) result = result && output_stat(statptr);
-    result = result && obuf_putfn(&out, filename, striplen);
-    if (list_flags) result = result && output_flags(statptr);
-    result = result && obuf_puts(&out, CRLF);
-    if (!result) {
-      close_out_connection(&out);
-      return respond_bytes(426, "Listing aborted", out.io.offset, 1);
+    line.len = 0;
+    if ((list_long && !str_catstat(&line, statptr))
+	|| !str_catfn(&line, filename, striplen)
+	|| (list_flags && !str_catflags(&line, statptr))
+	|| !str_cats(&line, CRLF)) {
+      close(out);
+      return respond_xferresult(-1, bytes_out, 1);
+    }
+    if ((result = send_line(out, &line)) != 0)
+      return respond_xferresult(result, bytes_out, 1);
+    bytes_out += line.len;
+  }
+  if (send_used > 0) {
+    if ((result = netwrite(out, send_buf, send_used, timeout * 1000)) != 0) {
+      socket_uncork(out);
+      close(out);
+      return respond_xferresult(result, bytes_out, 1);
     }
   }
-  if (!close_out_connection(&out))
-    return respond_bytes(426, "Listing aborted", out.io.offset, 1);
-  return respond_bytes(226, "Listing complete", out.io.offset, 1);
+  return respond_xferresult(close(out), bytes_out, 1);
 }
 
 static int list_dir()
 {
   long count;
-  if (!path_merge(&fullpath, "*") ||
-      (count = path_match(fullpath.s+1, &entries, list_options)) == -1)
-    return respond_internal_error();
-  return list_entries(count, str_findlast(&fullpath, '/'));
-}
+  DIR* dir;
+  direntry* entry;
 
-static int list_cwd()
-{
-  if (!str_copy(&fullpath, &cwd))
+  /* Turning "/" into "/." simplifies several other steps below. */
+  if (fullpath.len == 1)
+    if (!str_copys(&fullpath, "/."))
+      return respond_internal_error();
+
+  if ((dir = opendir(fullpath.s+1)) == 0)
+    return respond_syserr(550, "Could not open directory");
+  entries.len = 0;
+  count = 0;
+  while ((entry = readdir(dir)) != 0) {
+    /* Always skip the "." and ".." entries.  Skip all other ".*"
+     * entries if nodotfiles was set. */
+    if (entry->d_name[0] != '.'
+	|| (!(entry->d_name[1] == 0
+	      || (entry->d_name[1] == '.' && entry->d_name[2] == 0))
+	    && (list_options & PATH_MATCH_DOTFILES) != 0)) {
+      /* Add the entry to the list, skipping the "/" separator if none
+       * is needed */
+      if (!str_catb(&entries, fullpath.s+1, fullpath.len-1)
+	  || !str_catc(&entries, '/')
+	  || !str_catb(&entries, entry->d_name, strlen(entry->d_name)+1)) {
+	closedir(dir);
+	return respond_internal_error();
+      }
+      ++count;
+    }
+  }
+  closedir(dir);
+  if (!str_sort(&entries, 0, count, 0))
     return respond_internal_error();
-  return list_dir();
+  return list_entries(count, fullpath.len);
 }
 
 static int handle_listing(int longfmt)
@@ -253,21 +308,25 @@ static int handle_listing(int longfmt)
     }
   }
 
-  if (!req_param || !*req_param) req_param = ".";
+  if (!req_param || !*req_param)
+    req_param = ".";
   
   if (path_contains(req_param, ".."))
     return respond(553, 1, "Paths containing '..' not allowed.");
 
   /* Prefix the requested path with CWD, and strip it after */
   if (!qualify_validate(req_param)) return 1;
-  if (fullpath.len == 1) return list_cwd();
+  if (fullpath.len == 1) return list_dir();
   if ((count = path_match(fullpath.s+1, &entries, list_options)) == -1)
     return respond_internal_error();
   striplen = (req_param[0] == '/') ? -1 : (cwd.len == 1) ? 0 : cwd.len;
   
-  if (count == 0)
-    count = -1;
-  else if (count == 1) {
+  if (count <= 1) {
+    /* If no entries were matched, try the literal path. */
+    if (count == 0)
+      if (!str_copyb(&entries, fullpath.s+1, fullpath.len))
+	return respond_internal_error();
+
     if (stat(entries.s, &statbuf) == -1) {
       if (errno == EEXIST)
 	result = respond(550, 1, "File or directory does not exist.");
