@@ -1,5 +1,5 @@
 /* respond.c - twoftpd routines for responding to clients
- * Copyright (C) 2001  Bruce Guenter <bruceg@em.ca>
+ * Copyright (C) 2008  Bruce Guenter <bruce@untroubled.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,6 +18,7 @@
 #include <errno.h>
 #include <string.h>
 #include <iobuf/iobuf.h>
+#include <systime.h>
 #include "twoftpd.h"
 #include "log.h"
 
@@ -81,13 +82,49 @@ int respond(unsigned code, int final, const char* msg)
     respond_end();
 }
 
-int respond_bytes(unsigned code,
-		  const char* msg, unsigned long bytes, int sent)
+static struct timeval start;
+
+void respond_start_xfer(void)
 {
+  gettimeofday(&start, 0);
+}
+
+static const char* scales[] = { "", "k", "M", "G", "T", 0 };
+
+static int respond_bytes(unsigned code,
+			 const char* msg, unsigned long bytes, int sent)
+{
+  struct timeval end;
+  unsigned long rate;
+  int scaleno;
+  gettimeofday(&end, 0);
+  rate = bytes / (end.tv_sec-start.tv_sec +
+		  (end.tv_usec-start.tv_usec)/1000000.0);
+  for (scaleno = 0;
+       rate > 10000 && scales[scaleno+1] != 0;
+       ++scaleno, rate /= 1024)
+    ;
   return respond_start(code, 1) &&
     respond_str(msg) &&
     respond_str(" (") &&
     respond_uint(bytes) &&
-    respond_str(sent ? " bytes sent)." : " bytes received).") &&
+    respond_str(sent ? " bytes sent, " : " bytes received, ") &&
+    respond_uint(rate) &&
+    respond_str(scales[scaleno]) &&
+    respond_str("B/s).") &&
     respond_end();
 }
+
+int respond_xferresult(unsigned result, unsigned long bytes, int sent)
+{
+  switch (result) {
+  case 0:
+    return respond_bytes(226, "Transfer completed", bytes, sent);
+  case 1:
+    return respond_bytes(426, "Transfer timed out", bytes, sent);
+  case 2:
+    return respond_bytes(426, "Transfer interrupted", bytes, sent);
+  default:
+    return respond_bytes(451, "Transfer failed", bytes, sent);
+  }
+}
